from argparse import ArgumentParser
from getpass import getpass

from cc_agency.commons.conf import Conf
from cc_agency.commons.db import Mongo
from cc_agency.broker.auth import Auth

DESCRIPTION = 'Create a broker user to authenticate with the web API.'


def attach_args(parser):
    parser.add_argument(
        '-c', '--conf-file', action='store', type=str, metavar='CONF_FILE',
        help='CONF_FILE (yaml) as local path.'
    )


def main():
    parser = ArgumentParser(description=DESCRIPTION)
    attach_args(parser)
    args = parser.parse_args()

    return run(**args.__dict__)


def run(conf_file):
    conf = Conf(conf_file)
    mongo = Mongo(conf)
    auth = Auth(conf, mongo)

    print('You are in the process of creating a user account.')
    print('ATTENTION: an already existing user with the exact same username will be updated with new settings!')
    input('Hit [ENTER] to proceed...')

    username = input('Username: ')
    if not username:
        print('ABORT: username must not be empty.')
        exit(1)

    password = getpass('Password: ')
    if not password:
        print('ABORT: password must not be empty.')

    is_admin = input('Grant admin rights [y/N]: ')
    is_admin = is_admin.lower()
    is_admin = is_admin == 'yes' or is_admin == 'y'

    if is_admin:
        print('Admin privileges GRANTED!')
    else:
        print('Admin privileges NOT granted!')

    auth.create_user(username, password, is_admin=is_admin)
    print('Done!')



from time import time

import jsonschema
from flask import request, jsonify
from werkzeug.exceptions import Unauthorized, NotFound, BadRequest
from bson.objectid import ObjectId

from cc_agency.commons.schemas.callback import callback_schema
from cc_agency.commons.helper import batch_failure


def callback_routes(app, mongo, auth, conf, controller):
    @app.route('/callback/<batch_id>/<token>', methods=['GET'])
    def get_callback(batch_id, token):
        if not auth.verify_callback(batch_id, token):
            raise Unauthorized()

        batch = mongo.db['batches'].find_one(
            {'_id': ObjectId(batch_id), 'state': 'processing'}
        )
        if not batch:
            raise NotFound('Object not found.')

        experiment_id = batch['experimentId']

        experiment = mongo.db['experiments'].find_one(
            {'_id': ObjectId(experiment_id)}
        )

        result = {
            'redVersion': experiment['redVersion'],
            'cli': experiment['cli'],
            'inputs': batch['inputs'],
            'outputs': batch['outputs']
        }

        return jsonify(result)

    @app.route('/callback/<batch_id>/<token>', methods=['POST'])
    def post_callback(batch_id, token):
        if not auth.verify_callback(batch_id, token):
            raise Unauthorized()

        try:
            bson_id = ObjectId(batch_id)
        except Exception:
            raise BadRequest('Not a valid BSON ObjectId.')

        batch = mongo.db['batches'].find_one(
            {'_id': bson_id, 'state': 'processing'},
            {'attempts': 1, 'node': 1}
        )
        if not batch:
            raise NotFound('Object not found.')

        if not request.json:
            debug_info = 'Callback did not send CC-Agent data as JSON.'
            batch_failure(mongo, batch_id, debug_info, None, conf)
            raise BadRequest(debug_info)

        data = request.json

        try:
            jsonschema.validate(data, callback_schema)
        except Exception:
            debug_info = 'CC-Agent data sent by callback does not comply with jsonschema.'
            batch_failure(mongo, batch_id, debug_info, data, conf)
            raise BadRequest(debug_info)

        if data['state'] == 'failed':
            debug_info = 'Callback sent state "failed".'
            batch_failure(mongo, batch_id, debug_info, data, conf)

            return '', 200

        mongo.db['batches'].update(
            {'_id': bson_id},
            {
                '$set': {
                    'state': 'succeeded'
                },
                '$push': {
                    'history': {
                        'state': 'succeeded',
                        'time': time(),
                        'debugInfo': None,
                        'node': batch['node'],
                        'ccagent': data
                    }
                }
            }
        )

        controller.send_json({'destination': 'scheduler'})

        return '', 200



import json
from argparse import ArgumentParser
from subprocess import call
from time import sleep

from cc_agency.commons.conf import Conf

DESCRIPTION = 'Create a MongoDB admin user with read and write access, as specified in cc-agency configuration.'


def attach_args(parser):
    parser.add_argument(
        '-c', '--conf-file', action='store', type=str, metavar='CONF_FILE',
        help='CONF_FILE (yaml) as local path.'
    )
    parser.add_argument(
        '--host', action='store', type=str, metavar='HOST',
        help='Overwrite MongoDB host specified in CONF_FILE.'
    )


def main():
    parser = ArgumentParser(description=DESCRIPTION)
    attach_args(parser)
    args = parser.parse_args()

    return run(**args.__dict__)


def run(conf_file, host):
    conf = Conf(conf_file)

    host = host if host else conf.d['mongo'].get('host', 'localhost')
    port = conf.d['mongo'].get('port', 27017)
    db = conf.d['mongo']['db']
    username = conf.d['mongo']['username']
    password = conf.d['mongo']['password']

    data = {
        'pwd': password,
        'roles': [{
            'role': 'readWrite',
            'db': db
        }]
    }

    dumped = json.dumps(data)

    update_command = 'mongo --host {host} --port {port} --eval \'database = db.getSiblingDB("{db}"); database.updateUser("{username}", {dumped})\''.format(
        host=host,
        port=port,
        db=db,
        username=username,
        dumped=dumped
    )

    data['user'] = username

    dumped = json.dumps(data)

    create_command = 'mongo --host {host} --port {port} --eval \'database = db.getSiblingDB("{db}"); database.createUser({dumped})\''.format(
        host=host,
        port=port,
        db=db,
        dumped=dumped
    )

    for _ in range(10):
        code = call(update_command, shell=True)
        if code == 0:
            break
        else:
            code = call(create_command, shell=True)
            if code == 0:
                break
        sleep(1)



from cc_core.commons.schemas.common import pattern_key


_files_schema = {
    'type': 'object',
    'patternProperties': {
        pattern_key: {
            'type': 'object',
            'properties': {
                'isOptional': {'type': 'boolean'},
                'isArray': {'type': 'boolean'},
                'files': {
                    'type': 'array',
                    'items': {
                        'type': 'object',
                        'properties': {
                            'path': {'type': 'string'},
                            'size': {'type': 'number'},
                            'debugInfo': {'type': 'string'}
                        },
                        'additionalProperties': False,
                        'required': ['path', 'size', 'debugInfo']
                    }
                }
            },
            'additionalProperties': False,
            'required': ['isOptional', 'isArray', 'files']
        }
    },
    'additionalProperties': False
}

# TODO: Incomplete
callback_schema = {
    'type': 'object',
    'properties': {
        'state': {'enum': ['succeeded', 'failed']}
    },
    'additionalProperties': True,
    'required': ['state']
}



from argparse import ArgumentParser

from flask import Flask, jsonify, request
from werkzeug.exceptions import Unauthorized
import zmq

from cc_core.version import VERSION as CORE_VERSION
from cc_agency.version import VERSION as AGENCY_VERSION
from cc_agency.commons.conf import Conf
from cc_agency.commons.db import Mongo
from cc_agency.broker.auth import Auth
from cc_agency.broker.routes.red import red_routes
from cc_agency.broker.routes.nodes import nodes_routes
from cc_agency.broker.routes.callback import callback_routes
from cc_agency.broker.routes.auth import auth_routes


DESCRIPTION = 'CC-Agency Broker.'

app = Flask('broker')
application = app

parser = ArgumentParser(description=DESCRIPTION)
parser.add_argument(
    '-c', '--conf-file', action='store', type=str, metavar='CONF_FILE',
    help='CONF_FILE (yaml) as local path.'
)
args = parser.parse_args()

conf = Conf(args.conf_file)
mongo = Mongo(conf)
auth = Auth(conf, mongo)

context = zmq.Context()
controller = context.socket(zmq.PUSH)
controller.connect(conf.d['controller']['external_url'])


@app.route('/', methods=['GET'])
def get_root():
    return jsonify({'Hello': 'World'})


@app.route('/version', methods=['GET'])
def get_version():
    user = auth.verify_user(request.authorization)
    if not user:
        raise Unauthorized()

    return jsonify({
        'agencyVersion': AGENCY_VERSION,
        'coreVersion': CORE_VERSION
    })


red_routes(app, mongo, auth, controller)
nodes_routes(app, mongo, auth)
callback_routes(app, mongo, auth, conf, controller)
auth_routes(app, auth, conf)

controller.send_json({'destination': 'scheduler'})



from os import urandom
from queue import Queue
from threading import Thread
from time import time
from traceback import format_exc

import docker
from bson.objectid import ObjectId

from cc_agency.commons.helper import generate_secret, create_kdf, batch_failure


class ClientProxy:
    def __init__(self, node_name, conf, mongo):
        self._node_name = node_name
        self._conf = conf
        self._mongo = mongo

        node_conf = conf.d['controller']['docker']['nodes'][node_name]
        self._base_url = node_conf['base_url']
        self._tls = False
        if 'tls' in node_conf:
            self._tls = docker.tls.TLSConfig(**node_conf['tls'])

        self._external_url = conf.d['broker']['external_url'].rstrip('/')

        self._action_q = None
        self._client = None
        self._online = None

        node = {
            'nodeName': node_name,
            'state': None,
            'history': [],
            'ram': None,
            'cpus': None
        }

        bson_node_id = self._mongo.db['nodes'].insert_one(node).inserted_id
        self._node_id = str(bson_node_id)

        try:
            self._client = docker.DockerClient(base_url=self._base_url, tls=self._tls, version='auto')
            ram, cpus = self._info()
        except Exception:
            self._set_offline(format_exc())
            return

        self._set_online(ram, cpus)
        self._action_q = Queue()
        Thread(target=self._action_loop).start()

    def _set_online(self, ram, cpus):
        print('Node online:', self._node_name)

        self._online = True
        bson_node_id = ObjectId(self._node_id)
        self._mongo.db['nodes'].update_one(
            {'_id': bson_node_id},
            {
                '$set': {
                    'state': 'online',
                    'ram': ram,
                    'cpus': cpus
                },
                '$push': {
                    'history': {
                        'state': 'online',
                        'time': time(),
                        'debugInfo': None
                    }
                }
            }
        )

    def _set_offline(self, debug_info):
        print('Node offline:', self._node_name)
        timestamp = time()

        self._online = False
        bson_node_id = ObjectId(self._node_id)
        self._mongo.db['nodes'].update_one(
            {'_id': bson_node_id},
            {
                '$set': {'state': 'offline'},
                '$push': {
                    'history': {
                        'state': 'offline',
                        'time': timestamp,
                        'debugInfo': debug_info
                    }
                }
            }
        )

        # change state of assigned batches
        cursor = self._mongo.db['batches'].find(
            {'node': self._node_name, 'state': 'processing'},
            {'_id': 1}
        )

        for batch in cursor:
            bson_id = batch['_id']
            batch_id = str(bson_id)
            debug_info = 'Node offline: {}'.format(self._node_name)
            batch_failure(self._mongo, batch_id, debug_info, None, self._conf)

    def _info(self):
        info = self._client.info()
        ram = info['MemTotal'] // (1024 * 1024)
        cpus = info['NCPU']
        return ram, cpus

    def _batch_containers(self, status):
        batch_containers = {}

        if not self._online:
            return batch_containers

        containers = self._client.containers.list(all=True, limit=-1, filters={'status': status})

        for c in containers:
            try:
                ObjectId(c.name)
                batch_containers[c.name] = c
            except:
                pass

        return batch_containers

    def _fail_processing_batches_without_assigned_container(self):
        running_containers = self._batch_containers(None)

        cursor = self._mongo.db['batches'].find(
            {'state': 'processing'},
            {'_id': 1}
        )

        for batch in cursor:
            bson_id = batch['_id']
            batch_id = str(bson_id)

            if batch_id not in running_containers:
                debug_info = 'No container assigned.'
                batch_failure(self._mongo, batch_id, debug_info, None, self._conf)

    def _remove_cancelled_containers(self):
        running_containers = self._batch_containers('running')

        cursor = self._mongo.db['batches'].find(
            {
                '_id': {'$in': [ObjectId(_id) for _id in running_containers]},
                'state': 'cancelled'
            },
            {'_id': 1}
        )
        for batch in cursor:
            bson_id = batch['_id']
            batch_id = str(bson_id)

            c = running_containers[batch_id]
            c.remove(force=True)

    def _remove_exited_containers(self):
        exited_containers = self._batch_containers('exited')

        cursor = self._mongo.db['batches'].find(
            {'_id': {'$in': [ObjectId(_id) for _id in exited_containers]}},
            {'state': 1}
        )
        for batch in cursor:
            bson_id = batch['_id']
            batch_id = str(bson_id)

            c = exited_containers[batch_id]
            debug_info = c.logs().decode('utf-8')
            c.remove()

            if batch['state'] not in ['succeeded', 'failed', 'cancelled']:
                batch_failure(self._mongo, batch_id, debug_info, None, self._conf)

    def inspect_offline_node(self):
        try:
            self._client = docker.DockerClient(base_url=self._base_url, tls=self._tls, version='auto')
            ram, cpus = self._info()
            self._inspect()
        except Exception:
            return

        self._set_online(ram, cpus)
        self._action_q = Queue()
        Thread(target=self._action_loop).start()

    def _inspect(self):
        print('Node inspection:', self._node_name)

        core_image_conf = self._conf.d['controller']['docker']['core_image']
        image = core_image_conf['url']
        auth = core_image_conf.get('auth')
        command = 'ccagent connected {} --inspect'.format(self._external_url)
        disable_pull = self._conf.d['controller']['docker']['core_image'].get('disable_pull', False)

        if not disable_pull:
            self._client.images.pull(image, auth_config=auth)

        self._client.containers.run(
            image,
            command,
            user='1000:1000',
            remove=True
        )

    def put_action(self, data):
        self._action_q.put(data)

    def _action_loop(self):
        while True:
            data = self._action_q.get()

            if 'action' not in data:
                continue

            action = data['action']

            inspect = False

            if action == 'run_batch_container':
                try:
                    self._run_batch_container(batch_id=data['batch_id'])
                except Exception:
                    inspect = True
                    self._run_batch_container_failure(data['batch_id'], format_exc())

            elif action == 'pull_image':
                try:
                    self._pull_image(image=data['url'], auth=data.get('auth'))
                except:
                    inspect = True
                    batch_ids = data['required_by']
                    self._pull_image_failure(format_exc(), batch_ids)

            elif action == 'clean_up':
                try:
                    self._remove_cancelled_containers()
                    self._remove_exited_containers()
                    self._fail_processing_batches_without_assigned_container()
                except:
                    inspect = True

            if inspect:
                try:
                    self._inspect()
                except Exception:
                    self._set_offline(format_exc())
                    self._action_q = None
                    self._client = None
                    self._remove_cancelled_containers()
                    self._remove_exited_containers()
                    self._fail_processing_batches_without_assigned_container()

            if not self._online:
                return

    def _run_batch_container(self, batch_id):
        batch = self._mongo.db['batches'].find_one(
            {'_id': ObjectId(batch_id), 'state': 'processing'},
            {'experimentId': 1}
        )
        if not batch:
            print('Batch "{}" not found for processing.'.format(batch_id))
            return

        experiment_id = batch['experimentId']
        experiment = self._mongo.db['experiments'].find_one(
            {'_id': ObjectId(experiment_id)},
            {
                'container.settings.image.url': 1,
                'container.settings.ram': 1,
                'execution.settings.outdir': 1}
        )
        image = experiment['container']['settings']['image']['url']

        token = generate_secret()
        salt = urandom(16)
        kdf = create_kdf(salt)

        self._mongo.db['callback_tokens'].insert_one({
            'batch_id': batch_id,
            'salt': salt,
            'token': kdf.derive(token.encode('utf-8')),
            'timestamp': time()
        })

        command = 'ccagent connected {}/callback/{}/{}'.format(self._external_url, batch_id, token)

        if 'execution' in experiment:
            outdir = experiment['execution']['settings'].get('outdir')
            if outdir:
                command = '{} --outdir {}'.format(command, outdir)

        ram = experiment['container']['settings']['ram']
        mem_limit = '{}m'.format(ram)

        self._client.containers.run(
            image,
            command,
            name=batch_id,
            user='1000:1000',
            remove=False,
            detach=True,
            mem_limit=mem_limit,
            memswap_limit=mem_limit
        )

    def _run_batch_container_failure(self, batch_id, debug_info):
        batch_failure(self._mongo, batch_id, debug_info, None, self._conf)

    def _pull_image(self, image, auth):
        self._client.images.pull(image, auth_config=auth)

    def _pull_image_failure(self, debug_info, batch_ids):
        for batch_id in batch_ids:
            batch_failure(self._mongo, batch_id, debug_info, None, self._conf)

    def _remove_batch_container(self, batch_id):
        pass



from os import urandom
from binascii import hexlify
from time import time

from flask import request
from bson.objectid import ObjectId
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.hashes import SHA256
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC


def get_ip():
    headers = ['HTTP_X_FORWARDED_FOR', 'HTTP_X_REAL_IP', 'REMOTE_ADDR']
    ip = None
    for header in headers:
        ip = request.environ.get(header)
        if ip:
            break
    if not ip:
        ip = '127.0.0.1'
    return ip


def generate_secret():
    return hexlify(urandom(24)).decode('utf-8')


def create_kdf(salt):
    return PBKDF2HMAC(
        algorithm=SHA256(),
        length=32,
        salt=salt,
        iterations=100000,
        backend=default_backend()
    )


def batch_failure(mongo, batch_id, debug_info, ccagent, conf):
    bson_id = ObjectId(batch_id)

    batch = mongo.db['batches'].find_one(
        {'_id': bson_id},
        {'attempts': 1, 'node': 1}
    )

    timestamp = time()
    attempts = batch['attempts']
    node_name = batch['node']

    new_state = 'registered'
    new_node = None

    attempts_to_fail = conf.d['controller']['scheduling']['attempts_to_fail']

    if attempts >= attempts_to_fail:
        new_state = 'failed'
        new_node = node_name

    mongo.db['batches'].update(
        {'_id': bson_id},
        {
            '$set': {
                'state': new_state,
                'node': new_node
            },
            '$push': {
                'history': {
                    'state': new_state,
                    'time': timestamp,
                    'debugInfo': debug_info,
                    'node': new_node,
                    'ccagent': ccagent
                }
            }
        }
    )


def str_to_bool(s):
    if isinstance(s, str) and s.lower() in ['1', 'true']:
        return True
    return False



from os import urandom
from time import time

from cc_agency.commons.helper import generate_secret, get_ip, create_kdf


class Auth:
    def __init__(self, conf, mongo):
        self._num_login_attempts = conf.d['broker']['auth']['num_login_attempts']
        self._block_for_seconds = conf.d['broker']['auth']['block_for_seconds']
        self._tokens_valid_for_seconds = conf.d['broker']['auth']['tokens_valid_for_seconds']

        self._mongo = mongo

    def create_user(self, username, password, is_admin):
        salt = urandom(16)
        kdf = create_kdf(salt)
        user = {
            'username': username,
            'password': kdf.derive(password.encode('utf-8')),
            'salt': salt,
            'is_admin': is_admin
        }
        self._mongo.db['users'].update_one({'username': username}, {'$set': user}, upsert=True)

    def verify_user(self, auth):
        if not auth:
            return None

        username = auth.username
        password = auth.password

        user = self._mongo.db['users'].find_one({'username': username})
        user['verified_by_credentials'] = False

        if not user:
            return None

        salt = user['salt']
        del user['salt']

        if not user:
            return None

        if self._is_blocked_temporarily(user):
            return None

        if self._verify_user_by_token(user, password):
            return user

        if self._verify_user_by_credentials(user, password, salt):
            user['verified_by_credentials'] = True
            return user

        self._add_block_entry(user)
        return None

    def _is_blocked_temporarily(self, user):
        self._mongo.db['block_entries'].delete_many({'timestamp': {'$lt': time() - self._block_for_seconds}})
        block_entries = list(self._mongo.db['block_entries'].find({'username': user['username']}))

        if len(block_entries) > self._num_login_attempts:
            return True

        return False

    def _add_block_entry(self, username):
        self._mongo.db['block_entries'].insert_one({
            'username': username,
            'timestamp': time()
        })
        print('Unverified login attempt: added block entry!')

    def issue_token(self, user):
        salt = urandom(16)
        kdf = create_kdf(salt)
        token = generate_secret()
        self._mongo.db['tokens'].insert_one({
            'username': user['username'],
            'ip': get_ip(),
            'salt': salt,
            'token': kdf.derive(token.encode('utf-8')),
            'timestamp': time()
        })
        return token

    def _verify_user_by_token(self, user, token):
        self._mongo.db['tokens'].delete_many({'timestamp': {'$lt': time() - self._tokens_valid_for_seconds}})
        cursor = self._mongo.db['tokens'].find(
            {'username': user['username'], 'ip': get_ip()},
            {'token': 1, 'salt': 1}
        )
        for c in cursor:
            try:
                kdf = create_kdf(c['salt'])
                kdf.verify(token.encode('utf-8'), c['token'])
                return True
            except:
                pass

        return False

    def verify_callback(self, batch_id, token):
        self._mongo.db['callback_tokens'].delete_many({'timestamp': {'$lt': time() - self._tokens_valid_for_seconds}})
        cursor = self._mongo.db['callback_tokens'].find(
            {'batch_id': batch_id},
            {'token': 1, 'salt': 1}
        )
        for c in cursor:
            try:
                kdf = create_kdf(c['salt'])
                kdf.verify(token.encode('utf-8'), c['token'])
                return True
            except:
                pass

        return False

    @staticmethod
    def _verify_user_by_credentials(user, password, salt):
        kdf = create_kdf(salt)
        try:
            kdf.verify(password.encode('utf-8'), user['password'])
        except:
            return False

        return True



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['cc_agency',
 'cc_agency.broker',
 'cc_agency.broker.routes',
 'cc_agency.commons',
 'cc_agency.commons.schemas',
 'cc_agency.controller',
 'cc_agency.tools',
 'cc_agency.tools.create_broker_user',
 'cc_agency.tools.create_db_user',
 'cc_agency.tools.drop_db_collections']

package_data = \
{'': ['*']}

install_requires = \
['cc-core>=5.3,<5.4',
 'cryptography>=2.2,<3.0',
 'docker>=3.4,<4.0',
 'flask>=1.0,<2.0',
 'jsonschema>=2.6,<3.0',
 'pymongo>=3.7,<4.0',
 'pyzmq>=17.0,<18.0',
 'ruamel.yaml>=0.15,<0.16']

entry_points = \
{'console_scripts': ['ccagency = cc_agency.tools.main:main',
                     'ccagency-controller = cc_agency.controller.main:main']}

setup_kwargs = {
    'name': 'cc-agency',
    'version': '5.3.2',
    'description': 'CC-Agency is part of the Curious Containers project. It connects to a cluster of docker-engines for the distributed execution of reproducible data-driven experiments defined in RED format.',
    'long_description': '# CC-Agency\n\nCC-Agency is part of the Curious Containers project. It connects to a cluster of docker-engines for the distributed execution of reproducible data-driven experiments defined in RED format.\n\nFor more information please refer to the Curious Containers [documentation](https://www.curious-containers.cc/).\n\n## Acknowledgements\n\nThe Curious Containers software is developed at [CBMI](https://cbmi.htw-berlin.de/) (HTW Berlin - University of Applied Sciences). The work is supported by the German Federal Ministry of Economic Affairs and Energy (ZIM project BeCRF, grant number KF3470401BZ4), the German Federal Ministry of Education and Research (project deep.TEACHING, grant number 01IS17056) and HTW Berlin Booster.\n',
    'author': 'Christoph Jansen',
    'author_email': 'Christoph.Jansen@htw-berlin.de',
    'url': 'https://www.curious-containers.cc/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.4,<4.0',
}


setup(**setup_kwargs)



from threading import Thread
from queue import Queue
from time import time, sleep

from bson.objectid import ObjectId

from cc_agency.controller.docker import ClientProxy
from cc_agency.commons.helper import create_kdf
from binascii import hexlify


class Scheduler:
    def __init__(self, conf, mongo):
        self._conf = conf
        self._mongo = mongo

        mongo.db['nodes'].drop()

        self._scheduling_q = Queue(maxsize=1)
        self._inspection_q = Queue(maxsize=1)
        self._voiding_q = Queue(maxsize=1)

        self._nodes = {
            node_name: ClientProxy(node_name, conf, mongo)
            for node_name
            in conf.d['controller']['docker']['nodes'].keys()
        }

        Thread(target=self._scheduling_loop).start()
        Thread(target=self._inspection_loop).start()
        Thread(target=self._voiding_loop).start()
        Thread(target=self._cron).start()

    def _cron(self):
        while True:
            batch = self._mongo.db['batches'].find_one(
                {'state': {'$nin': ['succeeded', 'failed', 'cancelled']}},
                {'_id': 1}
            )
            if batch:
                self.schedule()

            sleep(60)

    def schedule(self):
        try:
            self._scheduling_q.put_nowait(None)
        except:
            pass

    def _inspection_loop(self):
        while True:
            self._inspection_q.get()
            print('inspecting...')

            cursor = self._mongo.db['nodes'].find(
                {'state': 'offline'},
                {'nodeName': 1, 'state': 1}
            )

            threads = []

            for node in cursor:
                node_name = node['nodeName']
                client_proxy = self._nodes[node_name]
                t = Thread(client_proxy.inspect_offline_node)
                t.start()
                threads.append(t)

            for t in threads:
                t.join()

    @staticmethod
    def _void_recursively(data, allowed_section):
        if isinstance(data, dict):
            result = {}
            for key, val in data.items():
                if allowed_section and (key == 'password' or key.startswith('_')):
                    kdf = create_kdf('fixedsalt'.encode('utf-8'))
                    val_hash = kdf.derive(val.encode('utf-8'))
                    val_hash = hexlify(val_hash).decode('utf-8')
                    result[key] = '{{' + val_hash[-8:] + '}}'
                elif key in ['access', 'auth']:
                    result[key] = Scheduler._void_recursively(val, True)
                else:
                    result[key] = Scheduler._void_recursively(val, allowed_section)

            return result
        elif isinstance(data, list):
            return [Scheduler._void_recursively(val, allowed_section) for val in data]

        return data

    def _voiding_loop(self):
        while True:
            self._voiding_q.get()
            print('voiding...')

            # batches
            cursor = self._mongo.db['batches'].find(
                {
                    'state': {'$in': ['succeeded', 'failed', 'cancelled']},
                    'protectedKeysVoided': False
                }
            )

            for batch in cursor:
                bson_id = batch['_id']
                del batch['_id']

                data = Scheduler._void_recursively(batch, False)
                data['protectedKeysVoided'] = True
                self._mongo.db['batches'].update_one({'_id': bson_id}, {'$set': data})

            # experiments
            cursor = self._mongo.db['experiments'].find(
                {
                    'protectedKeysVoided': False
                }
            )

            for experiment in cursor:
                bson_id = experiment['_id']
                experiment_id = str(bson_id)

                cursor = self._mongo.db['batches'].aggregate([
                    {'$match': {'experimentId': experiment_id}},
                    {'$count': 'count'}
                ])
                all_count = list(cursor)[0]['count']

                cursor = self._mongo.db['batches'].aggregate([
                    {
                        '$match': {
                            'experimentId': experiment_id,
                            'state': {'$in': ['succeeded', 'failed', 'cancelled']}
                        }
                    },
                    {'$count': 'count'}
                ])
                cursor = list(cursor)
                if cursor:
                    finished_count = cursor[0]['count']

                    if all_count == finished_count:
                        data = Scheduler._void_recursively(experiment, False)
                        data['protectedKeysVoided'] = True
                        self._mongo.db['experiments'].update_one({'_id': bson_id}, {'$set': data})

    def _scheduling_loop(self):
        while True:
            self._scheduling_q.get()
            print('scheduling...')

            # inspect offline nodes
            try:
                self._inspection_q.put_nowait(None)
            except:
                pass

            # void protected keys
            try:
                self._voiding_q.put_nowait(None)
            except:
                pass

            # schedule
            self._schedule_batches()

            # clean up online nodes
            cursor = self._mongo.db['nodes'].find(
                {'state': 'online'},
                {'nodeName': 1}
            )

            for node in cursor:
                node_name = node['nodeName']
                client_proxy = self._nodes[node_name]
                client_proxy.put_action({'action': 'clean_up'})

    def _online_nodes(self):
        cursor = self._mongo.db['nodes'].find(
            {'state': 'online'},
            {'ram': 1, 'nodeName': 1}
        )

        nodes = list(cursor)
        node_names = [node['nodeName'] for node in nodes]

        cursor = self._mongo.db['batches'].find(
            {'node': {'$in': node_names}, 'state': 'processing'},
            {'experimentId': 1, 'node': 1}
        )
        batches = list(cursor)
        experiment_ids = list(set([ObjectId(b['experimentId']) for b in batches]))

        cursor = self._mongo.db['experiments'].find(
            {'_id': {'$in': experiment_ids}},
            {'container.settings.ram': 1}
        )
        experiments = {str(e['_id']): e for e in cursor}

        for node in nodes:
            used_ram = sum([
                experiments[b['experimentId']]['container']['settings']['ram']
                for b in batches
                if b['node'] == node['nodeName']
            ])

            node['freeRam'] = node['ram'] - used_ram
            node['scheduledImages'] = {}
            node['scheduledBatches'] = []

        return nodes

    def _schedule_batches(self):
        nodes = self._online_nodes()
        strategy = self._conf.d['controller']['scheduling']['strategy']
        timestamp = time()

        experiments = {}

        # select batch to be scheduled
        for batch in self._fifo():
            batch_id = str(batch['_id'])
            experiment_id = batch['experimentId']

            experiment = experiments.get(experiment_id)

            if not experiment:
                experiment = self._mongo.db['experiments'].find_one(
                    {'_id': ObjectId(experiment_id)},
                    {'container.settings': 1, 'execution.settings': 1}
                )
                experiments[experiment_id] = experiment

            ram = experiment['container']['settings']['ram']

            # select node
            possible_nodes = [node for node in nodes if node['freeRam'] >= ram]

            if len(possible_nodes) == 0:
                continue

            if strategy == 'spread':
                possible_nodes.sort(reverse=True, key=lambda n: n['freeRam'])
            elif strategy == 'binpack':
                possible_nodes.sort(reverse=False, key=lambda n: n['freeRam'])

            selected_node = possible_nodes[0]
            selected_node['freeRam'] -= ram

            # schedule image pull on selected node
            disable_pull = False
            if 'execution' in experiment:
                disable_pull = experiment['execution']['settings'].get('disablePull', False)

            if not disable_pull:
                image_data = [experiment['container']['settings']['image']['url']]
                auth = experiment['container']['settings']['image'].get('auth')
                if auth:
                    image_data += [auth['username'], auth['password']]
                image_data = tuple(image_data)

                if image_data not in selected_node['scheduledImages']:
                    selected_node['scheduledImages'][image_data] = []

                selected_node['scheduledImages'][image_data].append(batch_id)

            # schedule batch on selected node
            selected_node['scheduledBatches'].append(batch)

            # update batch data
            self._mongo.db['batches'].update_one(
                {'_id': batch['_id']},
                {
                    '$set': {
                        'state': 'processing',
                        'node': selected_node['nodeName']
                    },
                    '$push': {
                        'history': {
                            'state': 'processing',
                            'time': timestamp,
                            'debugInfo': None,
                            'node': selected_node['nodeName'],
                            'ccagent': None
                        }
                    },
                    '$inc': {
                        'attempts': 1
                    }
                }
            )

        # inform node ClientProxies
        for node in nodes:
            node_name = node['nodeName']
            client_proxy = self._nodes[node_name]

            for image, required_by in node['scheduledImages'].items():
                data = {
                    'action': 'pull_image',
                    'url': image[0],
                    'required_by': required_by
                }

                if len(image) == 3:
                    data['auth'] = {
                        'username': image[1],
                        'password': image[2]
                    }

                client_proxy.put_action(data)

            for batch in node['scheduledBatches']:
                batch_id = str(batch['_id'])

                data = {
                    'action': 'run_batch_container',
                    'batch_id': batch_id
                }

                client_proxy.put_action(data)

    def _fifo(self):
        cursor = self._mongo.db['batches'].aggregate([
            {'$match': {'state': 'registered'}},
            {'$sort': {'registrationTime': 1}},
            {'$project': {'experimentId': 1}}
        ])
        for b in cursor:
            yield b



from argparse import ArgumentParser
import zmq

from cc_core.version import VERSION as CORE_VERSION
from cc_agency.version import VERSION as AGENCY_VERSION
from cc_agency.commons.conf import Conf
from cc_agency.commons.db import Mongo
from cc_agency.controller.scheduler import Scheduler


DESCRIPTION = 'CC-Agency Controller'


def main():
    print('CC-Agency Version:', AGENCY_VERSION)
    print('CC-Core Version:', CORE_VERSION)

    parser = ArgumentParser(description=DESCRIPTION)
    parser.add_argument(
        '-c', '--conf-file', action='store', type=str, metavar='CONF_FILE',
        help='CONF_FILE (yaml) as local path.'
    )
    args = parser.parse_args()

    # Singletons
    conf = Conf(args.conf_file)
    mongo = Mongo(conf)
    scheduler = Scheduler(conf, mongo)

    context = zmq.Context()
    socket = context.socket(zmq.PULL)
    socket.bind('tcp://{}:{}'.format(
        conf.d['controller']['bind_host'],
        conf.d['controller']['bind_port']
    ))

    while True:
        data = socket.recv_json()

        if 'destination' not in data:
            continue

        destination = data['destination']
        if destination == 'scheduler':
            scheduler.schedule()



conf_schema = {
    'type': 'object',
    'properties': {
        'broker': {
            'type': 'object',
            'properties': {
                'external_url': {'type': 'string'},
                'auth': {
                    'type': 'object',
                    'properties': {
                        'num_login_attempts': {'type': 'integer'},
                        'block_for_seconds': {'type': 'integer'},
                        'tokens_valid_for_seconds': {'type': 'integer'}
                    },
                    'additionalProperties': False,
                    'required': ['num_login_attempts', 'block_for_seconds', 'tokens_valid_for_seconds']
                }
            },
            'additionalProperties': False,
            'required': ['external_url', 'auth']
        },
        'controller': {
            'type': 'object',
            'properties': {
                'external_url': {'type': 'string'},
                'bind_host': {'type': 'string'},
                'bind_port': {'type': 'integer'},
                'docker': {
                    'type': 'object',
                    'properties': {
                        'core_image': {
                            'type': 'object',
                            'properties': {
                                'url': {'type': 'string'},
                                'auth': {
                                    'type': 'object',
                                    'properties': {
                                        'username': {'type': 'string'},
                                        'password': {'type': 'string'}
                                    },
                                    'additionalProperties': False,
                                    'required': ['username', 'password']
                                },
                                'disable_pull': {'type': 'boolean'}
                            },
                            'additionalProperties': False,
                            'required': ['url']
                        },
                        'nodes': {
                            'type': 'object',
                            'patternProperties': {
                                '^[a-zA-Z0-9_-]+$': {
                                    'type': 'object',
                                    'properties': {
                                        'base_url': {'type': 'string'},
                                        'tls': {
                                            'type': 'object',
                                            'properties': {
                                                'verify': {'type': 'string'},
                                                'client_cert': {
                                                    'type': 'array',
                                                    'items': {'type': 'string'}
                                                },
                                                'assert_hostname': {'type': ['boolean', 'string']}
                                            },
                                            'additionalProperties': True
                                        }
                                    },
                                    'required': ['base_url'],
                                    'additionalProperties': False
                                }
                            },
                            'additionalProperties': False
                        }
                    },
                    'additionalProperties': False,
                    'required': ['nodes']
                },
                'scheduling': {
                    'type': 'object',
                    'properties': {
                        'strategy': {'enum': ['spread', 'binpack']},
                        'attempts_to_fail': {'type': 'integer'}
                    },
                    'additionalProperties': False,
                    'required': ['strategy', 'attempts_to_fail']
                }
            },
            'additionalProperties': False,
            'required': ['external_url', 'bind_host', 'bind_port', 'docker', 'scheduling']
        },
        'mongo': {
            'type': 'object',
            'properties': {
                'host': {'type': 'string'},
                'port': {'type': 'integer'},
                'db': {'type': 'string'},
                'username': {'type': 'string'},
                'password': {'type': 'string'}
            },
            'additionalProperties': False,
            'required': ['db', 'username', 'password']
        }
    },
    'additionalProperties': False,
    'required': ['broker', 'controller', 'mongo']
}
